import itertools
import numpy as np
import numba
import strax
import straxen

from straxen.plugins.defaults import DEFAULT_POSREC_ALGO, FAR_XYPOS_S2_TYPE, WIDE_XYPOS_S2_TYPE
from .peak_proximity import half_cauchy_pdf
from .peak_ambience import distance_in_xy, _quick_assign

export, __all__ = strax.exporter()


@export
class PeakShadow(strax.OverlapWindowPlugin):
    """This plugin can find and calculate the time & position shadow from previous peaks in time. It
    also gives the area and (x, y) of the previous peaks.

    References:
        * v0.1.5 reference: xenon:xenonnt:ac:prediction:shadow_ambience

    """

    __version__ = "0.2.0"

    depends_on = ("peak_basics", "peak_positions")
    provides = "peak_shadow"
    save_when = strax.SaveWhen.EXPLICIT

    shadow_threshold = straxen.URLConfig(
        default={"s1_time_shadow": 1e3, "s2_time_shadow": 1e4, "s2_position_shadow": 1e4},
        type=dict,
        track=True,
        help="Only take S1/S2s larger than this into account when calculating Shadow [PE]",
    )

    shadow_sum = straxen.URLConfig(
        default={"s1_time_shadow": False, "s2_time_shadow": True, "s2_position_shadow": False},
        type=dict,
        track=True,
        help="Whether the shadow should be summed up rather than taking the largest one",
    )

    shadow_merge_replace = straxen.URLConfig(
        default={"s1_time_shadow": False, "s2_time_shadow": False, "s2_position_shadow": False},
        type=dict,
        track=True,
        help="Whether replace the primary peak with the merged peak",
    )

    shadow_merge_dt = straxen.URLConfig(
        default={"s1_time_shadow": -1, "s2_time_shadow": -1, "s2_position_shadow": -1},
        type=dict,
        track=True,
        help="Merge primary peaks within this time window [ns]",
    )

    shadow_merge_dr = straxen.URLConfig(
        default={"s1_time_shadow": -1, "s2_time_shadow": -1, "s2_position_shadow": -1},
        type=dict,
        track=True,
        help="Merge primary peaks within this distance [cm]",
    )

    shadow_merge_last_center = straxen.URLConfig(
        default={"s1_time_shadow": False, "s2_time_shadow": False, "s2_position_shadow": False},
        type=dict,
        track=True,
        help="Merge primary peaks but keep the last center time",
    )

    shadow_pdf_ordered = straxen.URLConfig(
        default={"s1_time_shadow": False, "s2_time_shadow": False, "s2_position_shadow": False},
        type=dict,
        track=True,
        help="Whether position shadow should be ordered by the position correlation only",
    )

    default_reconstruction_algorithm = straxen.URLConfig(
        default=DEFAULT_POSREC_ALGO, help="default reconstruction algorithm that provides (x,y)"
    )

    shadow_time_window_backward = straxen.URLConfig(
        default=int(1e9),
        type=int,
        track=True,
        help="Search for peaks casting time & position shadow in this time window [ns]",
    )

    shadow_uncertainty_weights = straxen.URLConfig(
        default=True,
        type=bool,
        track=True,
        help="Whether to use uncertainty weights when merging peaks",
    )

    shadow_deltatime_exponent = straxen.URLConfig(
        default=-1.0, type=float, track=True, help="The exponent of delta t when calculating shadow"
    )

    shadow_sigma_and_baseline = straxen.URLConfig(
        default=[15.220, 0.036],
        type=list,
        track=True,
        help=(
            "Fitted position correlation sigma[cm*PE^0.5] and baseline[cm] using in position shadow"
        ),
    )

    def get_window_size(self):
        return (10 * self.shadow_time_window_backward, 0)

    def infer_dtype(self):
        s1_time_shadow_dtype = []
        s2_time_shadow_dtype = []
        s2_position_shadow_dtype = []
        nearest_dtype = []
        # We have time shadow(S2/dt) and position shadow(S2/dt*p(s))
        # previous S1 can only cast time shadow, previous S2 can cast both time & position shadow
        for key, dtype in zip(
            ["s1_time_shadow", "s2_time_shadow", "s2_position_shadow"],
            [s1_time_shadow_dtype, s2_time_shadow_dtype, s2_position_shadow_dtype],
        ):
            type_str, tp_desc, _ = key.split("_")
            dtype.append(
                (
                    (
                        f"Previous large {type_str} casted largest {tp_desc} shadow [PE/ns]",
                        f"shadow_{key}",
                    ),
                    np.float32,
                )
            )
            dtype.append(
                (
                    (
                        (
                            f"Time difference to the previous large {type_str} peak casting largest"
                            f" {tp_desc} shadow [ns]"
                        ),
                        f"dt_{key}",
                    ),
                    np.int64,
                )
            )
            # Only previous S2 peaks have (x,y)
            if "s2" in key:
                dtype.append(
                    (
                        (
                            f"X of previous large s2 peak casting largest {tp_desc} shadow [cm]",
                            f"x_{key}",
                        ),
                        np.float32,
                    )
                )
                dtype.append(
                    (
                        (
                            f"Y of previous large s2 peak casting largest {tp_desc} shadow [cm]",
                            f"y_{key}",
                        ),
                        np.float32,
                    )
                )
            # Only time shadow gives the nearest large peak
            if "time" in key:
                dtype.append(
                    (
                        (
                            f"The nearest previous large {type_str} [ns]",
                            f"nearest_{type_str}",
                        ),
                        np.float32,
                    )
                )
                dtype.append(
                    (
                        (
                            f"Time difference to the nearest previous large {type_str} [ns]",
                            f"nearest_dt_{type_str}",
                        ),
                        np.int64,
                    )
                )
        # Also record the PDF of HalfCauchy when calculating S2 position shadow
        s2_position_shadow_dtype.append(
            (
                ("PDF describing correlation to the previous large s2", "pdf_s2_position_shadow"),
                np.float32,
            )
        )

        dtype = (
            s1_time_shadow_dtype
            + s2_time_shadow_dtype
            + s2_position_shadow_dtype
            + nearest_dtype
            + strax.time_fields
        )
        return dtype

    def setup(self):
        super().setup()
        for key in ["s2_position_shadow", "s2_time_shadow", "s1_time_shadow"]:
            if (
                self.shadow_sum[key]
                and not self.shadow_merge_replace[key]
                and self.shadow_merge_dt[key] > 0
            ):
                raise ValueError(
                    f"shadow_sum is set to True, but shadow_merge_replace is set to False for {key}"
                )
        if (
            self.shadow_uncertainty_weights
            and self.default_reconstruction_algorithm != DEFAULT_POSREC_ALGO
        ):
            raise ValueError(
                "Using pos-rec uncertainty as (x, y) merging weights, "
                f"but default_reconstruction_algorithm is not set to {DEFAULT_POSREC_ALGO}"
            )

    @staticmethod
    def cluster_xy(peaks, indices, dr):
        """Find connectivity map of peaks in xy space."""
        clusters = []
        unvisited = set(indices)

        while unvisited:
            current = unvisited.pop()
            cluster = {current}
            to_check = {current}
            while to_check:
                checking = to_check.pop()
                dists = distance_in_xy(peaks[checking], peaks[list(unvisited)])
                close = {i for i, d in zip(unvisited, dists) if d <= dr}
                to_check.update(close)
                cluster.update(close)
                unvisited -= close
            clusters.append(list(cluster))
        clusters = [sorted(cluster) for cluster in clusters if len(cluster) > 1]
        return clusters

    @staticmethod
    def merge_peaks(peaks, weights, dt, dr, replace=True, last=True):
        """Merge peaks in time and space."""

        # Prepare the dtype for the merged peaks
        dtype = []
        needed_fileds = "time endtime center_time area x y".split()
        for d, n in zip(peaks.dtype.descr, peaks.dtype.names):
            if n in needed_fileds:
                dtype.append(d)
        dtype = np.dtype(dtype)

        # Copy the fields to a new array
        _peaks = np.zeros(len(peaks), dtype=dtype)
        for d in needed_fileds:
            _peaks[d] = peaks[d]

        # Identify indices of peaks mergable in time
        if dt > 0:
            t_clusters = [[0]]
            for i in range(len(peaks) - 1):
                if np.abs(peaks["center_time"][i + 1] - peaks["center_time"][i]) > dt:
                    t_clusters.append([i])
                else:
                    t_clusters[-1].append(i + 1)
            t_clusters = [cluster for cluster in t_clusters if len(cluster) > 1]
        else:
            t_clusters = []

        # Identify indices of peaks mergable also in space
        if dr > 0:
            txy_clusters = []
            for cluster in t_clusters:
                txy_clusters += PeakShadow.cluster_xy(peaks, cluster, dr)
        else:
            txy_clusters = t_clusters

        # Merge peaks
        merged_peaks = np.zeros(len(txy_clusters), dtype=dtype)
        for i, cluster in enumerate(txy_clusters):
            # Merge peaks
            merged_peaks[i]["time"] = peaks["time"][cluster].min()
            merged_peaks[i]["endtime"] = strax.endtime(peaks)[cluster].max()
            # center_time is by definition the average time of waveform
            if last:
                merged_peaks[i]["center_time"] = peaks["center_time"][cluster[-1]]
            else:
                merged_peaks[i]["center_time"] = np.average(
                    peaks["center_time"][cluster], weights=peaks["area"][cluster]
                )
            merged_peaks[i]["area"] = np.sum(peaks["area"][cluster])
            for xy in ["x", "y"]:
                if np.any(np.isinf(weights[cluster])):
                    merged_peaks[i][xy] = peaks[xy][cluster][np.isinf(weights[cluster])].mean()
                else:
                    merged_peaks[i][xy] = np.average(peaks[xy][cluster], weights=weights[cluster])

        # Replace peaks with merged peaks if replace is True
        if replace:
            for i, cluster in enumerate(txy_clusters):
                _peaks[cluster] = merged_peaks[i]
            _peaks = np.delete(
                _peaks,
                list(itertools.chain.from_iterable([cluster[1:] for cluster in txy_clusters])),
            )
        else:
            _peaks = np.insert(_peaks, [cluster[0] for cluster in txy_clusters], merged_peaks)
        _peaks = strax.sort_by_time(_peaks)

        # Return merged peaks
        return _peaks

    @property
    def shadowdtype(self):
        dtype = []
        dtype += [("shadow", np.float32), ("dt", np.int64), ("pdf", np.float32)]
        dtype += [("x", np.float32), ("y", np.float32)]
        dtype += [("nearest", np.float32), ("nearest_dt", np.int64)]
        return dtype

    def compute(self, peaks):
        argsort = strax.stable_argsort(peaks["center_time"])
        _peaks = peaks[argsort].copy()
        result = np.zeros(len(peaks), self.dtype)
        _quick_assign(argsort, result, self.compute_shadow(peaks, _peaks))
        return result

    def compute_shadow(self, peaks, current_peak):
        # 1. Define time window for each peak, we will find previous peaks within these time windows
        roi = np.zeros(len(current_peak), dtype=strax.time_fields)
        roi["time"] = current_peak["center_time"] - self.shadow_time_window_backward
        roi["endtime"] = current_peak["center_time"]

        # 2. Calculate S2 position shadow, S2 time shadow, and S1 time shadow
        # weights of the peaklets when calculating the weighted mean deviation in (x, y)
        posrec_algo = self.default_reconstruction_algorithm
        if self.shadow_uncertainty_weights:
            weights = 1 / peaks[f"position_contour_area_{posrec_algo}"]
        else:
            weights = peaks["area"] * peaks["area_fraction_top"]
        result = np.zeros(len(current_peak), self.dtype)
        for key in ["s2_position_shadow", "s2_time_shadow", "s1_time_shadow"]:
            if "s2" in key:
                # in very rare cases, high energy S2 canbe classified as type 20 and 22
                stype = [2, FAR_XYPOS_S2_TYPE, WIDE_XYPOS_S2_TYPE]
            else:
                # type 3 is also S1 xenon:xenonnt:lsanchez:som_sr1b_summary_note
                stype = [1, 3]
            mask_pre = np.isin(peaks["type"], stype) & (peaks["area"] > self.shadow_threshold[key])
            _peaks = self.merge_peaks(
                peaks[mask_pre],
                weights[mask_pre],
                dt=self.shadow_merge_dt[key],
                dr=self.shadow_merge_dr[key],
                replace=self.shadow_merge_replace[key],
                last=self.shadow_merge_last_center[key],
            )
            split_peaks = strax.touching_windows(_peaks, roi)
            array = np.zeros(len(current_peak), np.dtype(self.shadowdtype))
            strax.set_nan_defaults(array)

            # Initialization
            array["dt"] = self.shadow_time_window_backward
            array["nearest_dt"] = self.shadow_time_window_backward
            # The default value for shadow is set to be the lowest possible value
            if "time" in key:
                array["shadow"] = (
                    self.shadow_threshold[key] * array["dt"] ** self.shadow_deltatime_exponent
                )
            else:
                # Assign the position shadow to be zero
                array["shadow"] = 0
            array["pdf"] = 0

            # Calculating shadow, the Major of the plugin.
            # Only record the previous peak casting the largest shadow if is_sum is False
            is_position = "position" in key
            is_sum = self.shadow_sum[key]
            if is_position and is_sum:
                raise ValueError(
                    "Position shadow cannot be summed up, please set its shadow_sum to False"
                )
            if len(current_peak):
                self.peaks_shadow(
                    current_peak,
                    _peaks,
                    split_peaks,
                    self.shadow_deltatime_exponent,
                    array,
                    is_position=is_position,
                    is_sum=is_sum,
                    is_pdf=self.shadow_pdf_ordered[key],
                    sigmas=self.getsigma(self.shadow_sigma_and_baseline, current_peak["area"]),
                )

            # Fill results
            type_str = key.split("_")[0]
            names = ["shadow", "dt"]
            if "s2" in key:  # Only previous S2 peaks have (x,y)
                names += ["x", "y"]
            if "time" in key:  # Only time shadow gives the nearest large peak
                names += ["nearest", "nearest_dt"]
            for name in names:
                if "nearest" in name:
                    result[f"{name}_{type_str}"] = array[name]
                else:
                    result[f"{name}_{key}"] = array[name]

            if key == "s2_position_shadow":
                # HalfCauchy PDF when calculating S2 position shadow
                result["pdf_s2_position_shadow"] = array["pdf"]

        # 6. Set time and endtime for peaks
        result["time"] = current_peak["time"]
        result["endtime"] = strax.endtime(current_peak)
        return result

    @staticmethod
    @np.errstate(invalid="ignore")
    def getsigma(sigma_and_baseline, s2):
        # The parameter of HalfCauchy, which is a function of S2 area
        return sigma_and_baseline[0] / np.sqrt(s2) + sigma_and_baseline[1]

    @staticmethod
    @numba.njit
    def peaks_shadow(
        peaks,
        pre_peaks,
        touching_windows,
        exponent,
        result,
        is_position=False,
        is_sum=False,
        is_pdf=False,
        sigmas=None,
    ):
        """For each peak in peaks, check if there is a shadow-casting peak and check if it casts the
        largest shadow."""
        for p_i, (suspicious_peak, sigma) in enumerate(zip(peaks, sigmas)):
            # casting_peak is the previous large peak casting shadow
            # suspicious_peak is the suspicious peak which in shadow from casting_peak
            indices = touching_windows[p_i]
            for idx in range(indices[0], indices[1]):
                casting_peak = pre_peaks[idx]
                dt = suspicious_peak["center_time"] - casting_peak["center_time"]
                if dt <= 0:
                    continue
                # First we record the time difference to the nearest previous peak
                if dt < result["nearest_dt"][p_i]:
                    result["nearest_dt"][p_i] = dt
                    result["nearest"][p_i] = casting_peak["area"]
                # Calculate time shadow
                new_shadow = casting_peak["area"] * dt**exponent
                if is_position or is_pdf:
                    # Calculate HalfCauchy PDF
                    distance = distance_in_xy(suspicious_peak, casting_peak)
                    # If distance is NaN, set largest distance
                    distance = np.where(np.isnan(distance), 2 * straxen.tpc_r, distance)
                    new_pdf = half_cauchy_pdf(distance, sigma).item()
                else:
                    new_pdf = np.nan
                if is_position:
                    new_shadow *= new_pdf
                # Only the previous peak with largest shadow is recorded
                found = is_pdf and new_pdf > result["pdf"][p_i]
                found |= ~is_pdf and new_shadow > result["shadow"][p_i]
                if found:
                    result["shadow"][p_i] = new_shadow
                    result["pdf"][p_i] = new_pdf
                    result["x"][p_i] = casting_peak["x"]
                    result["y"][p_i] = casting_peak["y"]
                    result["dt"][p_i] = suspicious_peak["center_time"] - casting_peak["center_time"]
            if is_sum:
                sl = slice(indices[0], indices[1])
                dt = suspicious_peak["center_time"] - pre_peaks["center_time"][sl]
                mask = dt > 0
                if mask.sum() == 0:
                    continue
                ft = dt[mask] ** exponent
                result["shadow"][p_i] = np.sum(pre_peaks["area"][sl][mask] * ft)
                result["pdf"][p_i] = np.nan
                result["x"][p_i] = np.nan
                result["y"][p_i] = np.nan
                result["dt"][p_i] = mask.sum() * np.sum(ft) ** (1 / exponent)



import strax
import straxen
import numpy as np
import numba
import typing as ty
from immutabledict import immutabledict

export, __all__ = strax.exporter()


@export
class nVETOEvents(strax.OverlapWindowPlugin):
    """Plugin which computes the boundaries of veto events."""

    __version__ = "0.1.1"

    depends_on = "hitlets_nv"
    provides = "events_nv"
    data_kind = "events_nv"
    compressor = "zstd"
    events_seen = 0

    event_left_extension_nv = straxen.URLConfig(
        default=0, track=True, type=int, help="Extends event window this many [ns] to the left."
    )

    event_resolving_time_nv = straxen.URLConfig(
        default=300, track=True, type=int, help="Resolving time for window coincidence [ns]."
    )

    event_min_hits_nv = straxen.URLConfig(
        default=4,
        track=True,
        type=int,
        help="Minimum number of fully confined hitlets to define an event.",
    )

    channel_map = straxen.URLConfig(
        track=False,
        type=immutabledict,
        help="immutabledict mapping subdetector to (min, max) channel number",
    )

    def infer_dtype(self):
        self.name_event_number = "event_number_nv"
        self.channel_range = self.channel_map["nveto"]
        self.n_channel = (self.channel_range[1] - self.channel_range[0]) + 1
        return veto_event_dtype(self.name_event_number, self.n_channel)

    def get_window_size(self):
        return self.event_left_extension_nv + self.event_resolving_time_nv + 1

    def compute(self, hitlets_nv):
        events, hitlets_ids_in_event = find_veto_events(
            hitlets_nv,
            self.event_min_hits_nv,
            self.event_resolving_time_nv,
            self.event_left_extension_nv,
            event_number_key=self.name_event_number,
            n_channel=self.n_channel,
        )

        # Compute basic properties:
        if len(hitlets_ids_in_event):
            compute_nveto_event_properties(
                events, hitlets_nv, hitlets_ids_in_event, start_channel=self.channel_range[0]
            )

        # Get eventids:
        n_events = len(events)
        events[self.name_event_number] = np.arange(n_events) + self.events_seen
        self.events_seen += n_events
        return events


def veto_event_dtype(
    name_event_number: str = "event_number_nv",
    n_pmts: int = 120,
) -> list:
    dtype = []
    dtype += strax.time_fields  # because mutable
    dtype += [
        (("Veto event number in this dataset", name_event_number), np.int64),
        (("Total area of all hitlets in event [pe]", "area"), np.float32),
        (("Total number of hitlets in events", "n_hits"), np.int32),
        (("Total number of contributing channels", "n_contributing_pmt"), np.uint8),
        (("Area in event per channel [pe]", "area_per_channel"), np.float32, n_pmts),
        (
            (
                "Area weighted mean time of the event relative to the event start [ns]",
                "center_time",
            ),
            np.float32,
        ),
        (("Weighted variance of time [ns]", "center_time_spread"), np.float32),
        (
            ("Minimal amplitude-to-amplitude gap between neighboring hitlets [ns]", "min_gap"),
            np.int32,
        ),
        (
            ("Maximal amplitude-to-amplitude gap between neighboring hitlets [ns]", "max_gap"),
            np.int32,
        ),
    ]
    return dtype


@numba.njit(cache=True, nogil=True)
def compute_nveto_event_properties(
    events: np.ndarray,
    hitlets: np.ndarray,
    contained_hitlets_ids: np.ndarray,
    start_channel: int = 2000,
):
    """Computes properties of the neutron-veto events. Writes results directly to events.

    :param events: Events for which properties should be computed
    :param hitlets: hitlets which were used to build the events.
    :param contained_hitlets_ids: numpy array of the shape n x 2 which holds the indices of the
        hitlets contained in the corresponding event.
    :param start_channel: Integer specifying start channel, e.g. 2000 for nveto.

    """
    for e, (s_i, e_i) in zip(events, contained_hitlets_ids):
        hitlets_in_event = hitlets[s_i:e_i]
        event_area = np.sum(hitlets_in_event["area"])
        e["area"] = event_area
        e["n_hits"] = len(hitlets_in_event)
        e["n_contributing_pmt"] = len(np.unique(hitlets_in_event["channel"]))

        t = hitlets_in_event["time"] - hitlets_in_event[0]["time"]

        dt = np.diff(hitlets_in_event["time"] + hitlets_in_event["time_amplitude"])
        e["min_gap"] = np.min(dt)
        e["max_gap"] = np.max(dt)

        if event_area:
            e["center_time"] = np.sum(t * hitlets_in_event["area"]) / event_area
            if e["n_hits"] > 1 and e["center_time"]:
                w = hitlets_in_event["area"] / e["area"]  # normalized weights
                # Definition of variance
                e["center_time_spread"] = np.sqrt(np.sum(w * np.power(t - e["center_time"], 2)))
            else:
                e["center_time_spread"] = np.inf

        # Compute per channel properties:
        for hit in hitlets_in_event:
            ch = hit["channel"] - start_channel
            e["area_per_channel"][ch] += hit["area"]


@export
def find_veto_events(
    hitlets: np.ndarray,
    coincidence_level: int,
    resolving_time: int,
    left_extension: int,
    event_number_key: str = "event_number_nv",
    n_channel: int = 120,
) -> ty.Tuple[np.ndarray, np.ndarray]:
    """Function which find the veto events as a nfold concidence in a given resolving time window.
    All hitlets which touch the event window contribute.

    :param hitlets: Hitlets which shall be used for event creation.
    :param coincidence_level: int, coincidence level.
    :param resolving_time: int, resolving window for coincidence in ns.
    :param left_extension: int, left event extension in ns.
    :param event_number_key: str, field name for the event number
    :param n_channel: int, number of channels in detector.
    :return: events, hitelt_ids_per_event

    """
    # Find intervals which satisfy requirement:
    event_intervals = straxen.plugins.nveto_recorder.find_coincidence(
        hitlets,
        coincidence_level,
        resolving_time,
        left_extension,
    )

    # Find all hitlets which touch the coincidence windows:
    # (we cannot use fully_contained in here since some muon signals
    # may be larger than 300 ns)
    hitlets_ids_in_event = strax.touching_windows(hitlets, event_intervals)

    # For some rare cases long signals may touch two intervals, in that
    # case we merge the intervals in the subsequent function:
    hitlets_ids_in_event = _solve_ambiguity(hitlets_ids_in_event)

    # Now we can create the veto events:
    events = np.zeros(
        len(hitlets_ids_in_event), dtype=veto_event_dtype(event_number_key, n_channel)
    )
    _make_event(hitlets, hitlets_ids_in_event, events)
    return events, hitlets_ids_in_event


@numba.njit(cache=True, nogil=False)
def _solve_ambiguity(contained_hitlets_ids: np.ndarray) -> np.ndarray:
    """Function which solves the ambiguity if a single hitlets overlaps with two event intervals.

    This can happen for muon signals which have a long tail, since we define the coincidence window
    as a fixed window. Hence those tails can extend beyond the fixed window.

    """
    res = np.zeros(contained_hitlets_ids.shape, dtype=contained_hitlets_ids.dtype)

    if not len(res):
        # Return empty result
        return res

    offset = 0
    start_i, end_i = contained_hitlets_ids[0]
    for e_i, ids in enumerate(contained_hitlets_ids[1:]):
        if end_i > ids[0]:
            # Current and next interval overlap so just updated the end
            # index.
            end_i = ids[1]
        else:
            # They do not overlap store indices:
            res[offset] = [start_i, end_i]
            offset += 1
            # Init next interval:
            start_i, end_i = ids

    # Last event:
    res[offset, :] = [start_i, end_i]
    offset += 1
    return res[:offset]


@numba.njit(cache=True, nogil=True)
def _make_event(hitlets: np.ndarray, hitlet_ids: np.ndarray, res: np.ndarray):
    """Function which sets veto event time and endtime."""
    for ei, ids in enumerate(hitlet_ids):
        hit = hitlets[ids[0] : ids[1]]
        res[ei]["time"] = hit[0]["time"]
        endtime = np.max(strax.endtime(hit))
        res[ei]["endtime"] = endtime



import param
from immutabledict import immutabledict
import numpy as np
import pandas as pd
import holoviews as hv
import straxen

# Define function which plots points with radius defined in data space:
# Taken from:
# https://stackoverflow.com/questions/60361810/how-do-i-set-the-scatter-circle-radius-in-holoviews
from holoviews.element.chart import Chart
from holoviews.plotting.bokeh import PointPlot


class PlotPMTArrayTPC:
    """Class to plot TPC PMT array.

    :param gains: If a list of gain values are supplied PMTs which gain are set to zero are
        displayed in gray.
    :param top_pmt_array: Boolean whether top or bottom PMT array should be plotted.

    """

    def __init__(
        self,
        gains=None,
        top_pmt_array=True,
        opts=immutabledict(),
    ):
        self.is_top_pmt_array = top_pmt_array

        self._init_start_end_index()
        self._init_pmt_array(gains)
        self._init_common_options(opts)

    def _init_start_end_index(self):
        """Defines index-range of channels to be used for the plot.

        Depends on whether the top or bottom array should be plotted.

        """
        self.pmt_start = 0 if self.is_top_pmt_array else straxen.n_top_pmts + 1
        self.pmt_end = straxen.n_top_pmts if self.is_top_pmt_array else straxen.n_tpc_pmts

    def _init_pmt_array(self, gains):
        """Initalized a dataframe which will contain the Top PMT pattern array."""
        _pmts = pd.DataFrame()
        _pmts["channel"] = np.arange(0, straxen.n_tpc_pmts, 1)
        _pmts["x"] = straxen.pmt_positions().loc[:, "x"].values
        _pmts["y"] = straxen.pmt_positions().loc[:, "y"].values
        _pmts["gains"] = np.ones(straxen.n_tpc_pmts)
        if np.any(gains):
            if len(gains) != straxen.n_tpc_pmts:
                raise ValueError("Expected gains to have the same shape as number of PMTs in TPC!")
            _is_zero = gains == 0
            _pmts.loc[_is_zero, "gains"] = np.nan
        self.pmts = _pmts.loc[self.pmt_start : self.pmt_end - 1, :]

    def _init_common_options(self, opt_dict):
        """Creates dictionary with common options required by PMT array plot."""
        if not isinstance(opt_dict, (dict, immutabledict)):
            raise ValueError("opts_dict must be a dictionary!")

        if self.is_top_pmt_array:
            title = "Top array top view"
        else:
            title = "Bottom array top view"

        r = straxen.tpc_r
        r *= 1.1
        default_settings = {
            "xlim": (-r, r),
            "ylim": (-r, r),
            "xlabel": "x [cm]",
            "ylabel": "y [cm]",
            "title": title,
        }
        self.settings = immutabledict({**default_settings, **opt_dict})

    @staticmethod
    def _plot_tpc(**opts):
        """Function which plots the TPC as a simple circle."""
        tpc = straxen.holoviews_utils.plot_tpc_circle(straxen.tpc_r)
        tpc.opts(**opts)
        return tpc

    def _plot_top_array(self, peak_data, label, **opts):
        pmts = pd.DataFrame()
        pmts["area"] = peak_data["area_per_channel"][self.pmt_start : self.pmt_end]

        # Cops other static values:
        # (we cannot use self.pmts as plot input as new plots will also change prev.
        # plots due to same reference.)
        pmts["channel"] = self.pmts["channel"]
        pmts["x"] = self.pmts["x"]
        pmts["y"] = self.pmts["y"]
        pmts["gains"] = self.pmts["gains"]

        pmts["area"] /= pmts["gains"]
        pmt_array = Circle(
            pmts,
            kdims=[hv.Dimension("x", label="x [cm]")],
            vdims=[
                hv.Dimension("y", label="y [cm]"),
                hv.Dimension("area", label="Area [pe]"),
                hv.Dimension("channel", label="Channel"),
            ],
            label=label,
        )
        pmt_array.opts(
            color="area",
            cmap="viridis",
            line_color="black",
            radius=3 * 2.54 / 2,
            tools=["hover"],
            clim=(10, None),
            clipping_colors={"NaN": "gray"},
        )
        pmt_array.opts(**opts)
        return pmt_array

    def plot_pmt_array(self, peak, label="", **opts):
        """Function which plots the specified PMT array for the given peak.

        :param peak: Peak for which PMT array should be plotted. Must contain area per channel
            information.
        :param label: Plot abel to be used as title.
        :param opts: Option which can be supplied to a holovies plot. E.g. you can use logz=True for
            a logarithmic colorscale or colorbar=True for adding a colorbar.
        :return: Holoviews plot overlay.

        """
        # Allow to return empty plots required for event display if alt.
        if peak is None:
            return hv.Points(None)

        plot = self._plot_tpc()
        plot *= self._plot_top_array(peak, label=label).opts(**self.settings, **opts)
        return plot


class Circle(Chart):
    group = param.String(default="Circle", constant=True)
    size = param.Integer()


class CirclePlot(PointPlot):
    _plot_methods = dict(single="circle", batched="circle")

    style_opts = ["radius" if so == "size" else so for so in PointPlot.style_opts if so != "marker"]


hv.Store.register({Circle: CirclePlot}, "bokeh")



import numpy as np
import strax
import straxen

from straxen.plugins.defaults import DEFAULT_POSREC_ALGO

export, __all__ = strax.exporter()


@export
class PeakPositions(strax.MergeOnlyPlugin):
    """Merge the reconstructed algorithms of the different algorithms into a single one that can be
    used in Event Basics.

    Select one of the plugins to provide the 'x' and 'y' to be used further down the chain. Since we
    already have the information needed here, there is no need to wait until events to make the
    decision.

    Since the computation is trivial as it only combined the three input plugins, don't save this
    plugins output.

    """

    __version__ = "0.0.0"
    provides = "peak_positions"
    depends_on = (
        "peak_positions_cnf",
        "peak_positions_mlp",
    )
    save_when = strax.SaveWhen.NEVER

    default_reconstruction_algorithm = straxen.URLConfig(
        default=DEFAULT_POSREC_ALGO, help="default reconstruction algorithm that provides (x,y)"
    )

    def infer_dtype(self):
        dtype = strax.merged_dtype([self.deps[d].dtype_for(d) for d in sorted(self.depends_on)])
        dtype += [
            ("x", np.float32, "Reconstructed S2 X position (cm), uncorrected"),
            ("y", np.float32, "Reconstructed S2 Y position (cm), uncorrected"),
        ]
        return dtype

    def compute(self, peaks):
        result = {dtype: peaks[dtype] for dtype in peaks.dtype.names}
        algorithm = self.default_reconstruction_algorithm
        for xy in ("x", "y"):
            result[xy] = peaks[f"{xy}_{algorithm}"]
        return result



import os
import argparse
import pymongo
import hug
import socket
from datetime import datetime, timedelta
import pytz
import time
import threading

import strax
import straxen

st: strax.Context
max_load_mb = 1000
max_return_mb = 10


def load_context(name, extra_dirs=None):
    global st
    st = getattr(straxen.contexts, name)()
    for sf in st.storage:
        sf.readonly = True

    if extra_dirs is not None:
        for d in extra_dirs:
            if os.access(d, os.W_OK) is not True:
                raise IOError(f"No writing access to {d}. Check mountpoints ")
            st.storage.append(strax.DataDirectory(d, readonly=True))

    st.context_config["allow_incomplete"] = True
    st.context_config["forbid_creation_of"] = (
        "*",
        # In case someone has an old strax that doesn't support * here:
        "raw_records",
        "records",
        "peaklets",
        "peaks",
        "event_basics",
        "event_info",
    )


@hug.exception(Exception)
def handle_exception(exception):
    return {"error": str(exception)}


@hug.get("/get_data")
def get_data(
    run_id: hug.types.text,
    target: hug.types.text,
    max_n: hug.types.number = 1000,
    start: hug.types.float_number = None,
    end: hug.types.float_number = None,
    selection: hug.types.text = None,
):
    try:
        st
    except NameError:
        return "Context not loaded???"

    if len(run_id) != 6 and run_id.isdecimal():
        run_id = "%06d" % int(run_id)

    if max_n <= 0:
        max_n = None
    t0, _ = st.estimate_run_start_and_end(run_id, target)
    if start is None or end is None:
        time_range = None
    else:
        time_range = [t0 + int(start * int(1e9)), t0 + int(end * int(1e9))]

    md = st.get_metadata(run_id, target)  # don't catch exception, already nice
    if not md["chunks"]:
        raise ValueError(
            "No chunks available -- either the first chunk has "
            "yet to be written, or something is wrong."
        )
    full_run_bytes = st.size_mb(run_id, target) * int(1e6)

    if time_range is None:
        # Time range is the full run. Find the second range corresponding
        # to that (as best we can)
        if "chunks" not in md or not md["chunks"]:
            print(f"Run {run_id} has no chunks??")
        else:
            time_range = [md["chunks"][0]["start"], md["chunks"][-1]["end"]]

    if time_range is None or not max_n:
        # We have to load all the data
        load_bytes = full_run_bytes

    else:
        # Let's see how much we really need to load, given max_n.
        load_at_least_n = 0
        load_bytes = 0
        for chunk_info in md["chunks"]:
            if chunk_info["start"] > time_range[1] or chunk_info["end"] < time_range[0]:
                # None of this chunk will be loaded
                continue
            load_bytes += chunk_info["nbytes"]

            if chunk_info["start"] > time_range[0] and chunk_info["end"] < time_range[1]:
                # All of this chunk will be loaded
                load_at_least_n += chunk_info["n"]

            if load_at_least_n > max_n:
                # No point loading any more, we already exceeded max_n
                time_range[1] = chunk_info["end"]
                break

    if load_bytes > max_load_mb * int(1e6):
        raise ValueError(
            f"Cannot load {target} for run {run_id}: it would "
            f"take more than {max_load_mb} MB RAM to load."
        )

    x = st.get_array(run_id, target, time_range=time_range, selection=selection)
    if max_n:
        x = x[:max_n]
    size_mb = x.nbytes / int(1e6)
    if size_mb > max_return_mb:
        raise ValueError(
            f"Not converting {target} for run {run_id} "
            f"to json since the binary data is {size_mb:.1f} MB. "
            "Try lowering max_n."
        )

    return [dict(zip(row.dtype.names, row)) for row in x]


def now(plus=0):
    return datetime.now(pytz.utc) + timedelta(seconds=plus)


def _set_state():
    """Inform the bootstrax collection we're running microstrax."""
    # DAQ database
    daq_db_name = "daq"
    daq_uri = straxen.get_mongo_uri(
        header="rundb_admin",
        user_key="mongo_daq_username",
        pwd_key="mongo_daq_password",
        url_key="mongo_daq_url",
    )
    daq_client = pymongo.MongoClient(daq_uri)
    daq_db = daq_client[daq_db_name]
    bs_coll = daq_db["eb_monitor"]
    while True:
        microstrax_state = dict(
            host=socket.getfqdn(),
            pid=os.getpid(),
            time=now(),
            state="hosting microstrax",
            max_load_mb=args.max_load_mb,
            max_return_mb=args.max_return_mb,
            context=args.context,
            n_dirs=len(args.extra_dirs),
        )
        try:
            daq_db.command("ping")
        except Exception as timeout_error:
            print(f"Ran into {timeout_error}. Can be bad. Let's take a nap")
        bs_coll.insert_one(microstrax_state)
        time.sleep(9)


def set_state():
    """Open thread to set the state of microstrax."""
    update_thread = threading.Thread(name="Update_db", target=_set_state)
    update_thread.setDaemon(True)
    update_thread.start()


def main():
    load_context(args.context, extra_dirs=args.extra_dirs)
    set_state()

    while True:
        try:
            hug.API(__name__).http.serve(port=args.port)
        except pymongo.errors.NotMasterError:
            print("Ran into the infamous pymongo.errors.NotMasterError. Sleep for a sec")
            time.sleep(60)


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description="Start a microservice to return strax data as json"
    )
    parser.add_argument(
        "--context", default="xenonnt_online", help="Name of straxen context to use"
    )
    parser.add_argument("--port", default=8000, type=int, help="HTTP port to serve on")
    parser.add_argument("--max_load_mb", default=1000, type=int)
    parser.add_argument("--max_return_mb", default=10, type=int)
    parser.add_argument("--extra_dirs", nargs="*", help="Extra directories to look for data")
    args = parser.parse_args()

    max_load_mb = args.max_load_mb
    max_return_mb = args.max_return_mb

    main()



import numpy as np
import numpy.lib.recfunctions as rfn
import numba
from straxen.plugins.peaklets.peaklet_classification_vanilla import PeakletClassificationVanilla

import strax
import straxen

export, __all__ = strax.exporter()

__all__.extend(["som_additional_fields"])


som_additional_fields = [
    ("vanilla_type", np.int8, "Vanilla type of the peak(let)"),
    ("som_type", np.int8, "SOM type of the peak(let)"),
    ("som_sub_type", np.int32, "SOM subtype of the peak(let)"),
    ("loc_x_som", np.int16, "x location of the peak(let) in the SOM"),
    ("loc_y_som", np.int16, "y location of the peak(let) in the SOM"),
]


@export
class PeakletClassificationSOM(PeakletClassificationVanilla):
    """
    Self-Organizing Maps (SOM)
    https://xe1t-wiki.lngs.infn.it/doku.php?id=xenon:xenonnt:lsanchez:unsupervised_neural_network_som_methods
    https://xe1t-wiki.lngs.infn.it/doku.php?id=xenon:xenonnt:lsanchez:som_summary_note
    For peaklet classification. We this pluggin will provide 2 data types, the 'type' we are
    already familiar with, classifying peaklets as s1, s2 (using the new classification) or
    unknown (from the previous classification). As well as a new data type, SOM type, which
    will be assigned numbers based on the cluster in the SOM in which they are found. For
    each version I will make some documentation in the corrections repository explaining
    what I believe each cluster represents.

    This correction/plugin is currently on the testing phase, feel free to use it if you are
    curious or just want to test it or try it out but note this is note ready to be used in
    analysis.

    Here data will be classified into one of four types:
      Type 0: Unknown, these are all data that associated with being non-physical like
              dark counts or unknown signals that do not seem to come from interactions
              of interest in our detector.
      Type 1: S1 interactions produced by excited xenon molecules in the liquid xenon
      Type 2: S2 interactions produced when freed electrons in the LXe breach the liquid
              gas interface.
      Type 3: Gas events. S1 interactions that occur on the GXe instead of the liquid.
              Have S1 waveform shape, with higher AFT.
    """

    __version__ = "0.2.1"
    child_plugin = True

    dtype = (
        strax.peak_interval_dtype
        + [
            ("type", np.int8, "Classification of the peak(let)"),
        ]
        + som_additional_fields
    )

    som_files = straxen.URLConfig(
        default="resource://xedocs://som_classifiers?attr=value"
        "&version=ONLINE&run_id=plugin.run_id&fmt=npy"
    )

    use_som_as_default = straxen.URLConfig(
        default=True,
        help=(
            "Boolean to indicate whether to use SOM "
            "or the straxen classifcation method for "
            "peaklet classification."
        ),
    )

    def setup(self):
        self.som_data = self.som_files
        self.som_weight_cube = self.som_files["weight_cube"]
        self.som_img = self.som_files["som_img"]
        self.som_norm_factors = self.som_files["norm_factors"]
        self.som_s1_array = self.som_files["s1_array"]
        self.som_s2_array = self.som_files["s2_array"]
        self.som_s3_array = self.som_files["s3_array"]
        self.som_s0_array = self.som_files["s0_array"]

    def compute(self, peaklets):
        # Current classification
        _peaklets_classifcation = super().compute(peaklets)

        peaklets_classifcation = np.zeros(len(_peaklets_classifcation), dtype=self.dtype)
        strax.set_nan_defaults(peaklets_classifcation)
        strax.copy_to_buffer(
            _peaklets_classifcation, peaklets_classifcation, "_copy_peaklets_information"
        )
        peaklets_classifcation["vanilla_type"] = _peaklets_classifcation["type"]
        peaklets_classifcation["som_sub_type"] = _peaklets_classifcation["type"]
        del _peaklets_classifcation

        # SOM classification
        _is_s1_or_s2 = np.isin(peaklets_classifcation["vanilla_type"], [1, 2])
        _peaklets = np.copy(peaklets[_is_s1_or_s2])
        _peaklets["type"] = peaklets_classifcation["type"][_is_s1_or_s2]

        som_sub_type, x_som, y_som = recall_populations(
            _peaklets,
            self.som_weight_cube,
            self.som_img,
            self.som_norm_factors,
            self.som_data,
        )
        peaklets_classifcation["som_type"][_is_s1_or_s2] = som_type_to_type(
            som_sub_type, self.som_s1_array, self.som_s2_array, self.som_s3_array, self.som_s0_array
        )
        peaklets_classifcation["som_sub_type"][_is_s1_or_s2] = som_sub_type
        peaklets_classifcation["loc_x_som"][_is_s1_or_s2] = x_som
        peaklets_classifcation["loc_y_som"][_is_s1_or_s2] = y_som
        if self.use_som_as_default:
            peaklets_classifcation["type"] = peaklets_classifcation["som_type"]
        else:
            peaklets_classifcation["type"] = peaklets_classifcation["vanilla_type"]

        return peaklets_classifcation


def recall_populations(dataset, weight_cube, som_cls_img, norm_factors, som_data):
    """Master function that should let the user provide a weightcube, a reference img as a np.array,
    a dataset and a set of normalization factors.

    In theory, if these 5 things are provided, this function should output
    the original data back with one added field with the name "som_sub_type"
    weight_cube:      SOM weight cube (3D array)
    som_cls_img:      SOM reference image as a numpy array
    dataset:          Data to preform the recall on (Should be peaklet level data)
    normfactos:       A set of 11 numbers to normalize the data so we can preform a recall

    """

    xdim, ydim, zdim = weight_cube.shape
    img_xdim, img_ydim, img_zdim = som_cls_img.shape
    unique_colors = np.unique(np.reshape(som_cls_img, [xdim * ydim, 3]), axis=0)
    # Checks that the reference image matches the weight cube
    assert (
        xdim == img_xdim
    ), f"Dimensions mismatch between SOM weight cube ({xdim}) and reference image ({img_xdim})"
    assert (
        ydim == img_ydim
    ), f"Dimensions mismatch between SOM weight cube ({ydim}) and reference image ({img_ydim})"

    assert all(dataset["type"] != 0), "Dataset contains unclassified peaklets"
    # Get the deciles representation of data for recall
    decile_transform_check = data_to_log_decile_log_area_aft(dataset, norm_factors)
    if "reduce_decile" in som_data.keys():
        if som_data["reduce_decile"]:
            decile_transform_check = decile_transform_check[:, 1:]
    # preform a recall of the dataset with the weight cube
    # assign each population color a number (can do from previous function)
    ref_map = generate_color_ref_map(som_cls_img, unique_colors, xdim, ydim)
    som_cls_array = np.full(len(dataset["area"]), np.nan)
    # Make new numpy structured array to save the SOM cls data
    data_with_SOM_cls = rfn.append_fields(dataset, "som_sub_type", som_cls_array)
    # preforms the recall and assigns som_sub_type label
    output_data, x_som, y_som = som_cls_recall(
        data_with_SOM_cls, decile_transform_check, weight_cube, ref_map
    )
    return output_data["som_sub_type"], x_som, y_som


def generate_color_ref_map(color_image, unique_colors, xdim, ydim):
    ref_map = np.zeros((xdim, ydim))
    for color in np.arange(len(unique_colors)):
        mask = np.all(np.equal(color_image, unique_colors[color, :]), axis=2)
        indices = np.argwhere(mask)  # generates a 2d mask
        for loc in np.arange(len(indices)):
            ref_map[indices[loc][0], indices[loc][1]] = color
    return ref_map


@export
def euclidean_dist(XA, XB):
    # mimicking scipy.spatial.distance.cdist when metric='euclidean'
    assert XA.shape[-1] == XB.shape[1], "Dimensions of points in XA and XB must match."
    return _euclidean_dist(XA, XB)


@numba.njit
def _euclidean_dist(XA, XB):
    nA, dA = XA.shape
    nB, dB = XB.shape
    distances = np.empty((nA, nB))
    for i in range(nA):
        for j in range(nB):
            dist = 0.0
            for k in range(dA):
                diff = XA[i, k] - XB[j, k]
                dist += diff * diff
            distances[i, j] = np.sqrt(dist)
    return distances


def som_cls_recall(array_to_fill, data_in_som_fmt, weight_cube, reference_map):
    som_xdim, som_ydim, _ = weight_cube.shape
    # for data_point in data_in_SOM_fmt:
    distances = euclidean_dist(weight_cube.reshape(-1, weight_cube.shape[-1]), data_in_som_fmt)
    w_neuron = np.argmin(distances, axis=0)
    x_idx, y_idx = np.unravel_index(w_neuron, (som_xdim, som_ydim))
    array_to_fill["som_sub_type"] = reference_map[x_idx, y_idx]
    return array_to_fill, x_idx, y_idx


def som_type_to_type(som_type, s1_array, s2_array, s3_array, s0_array):
    """
    Converts the SOM type into either S1 or S2 type (1, 2)
    som_type:    array with integers corresponding to the different SOM types
    s1_array:    array containing the number corresponding to the SOM types which should
                 be converted to S1's
    """
    som_type_copy = som_type.copy()
    som_type_copy[np.isin(som_type_copy, s1_array)] = 1234
    som_type_copy[np.isin(som_type_copy, s2_array)] = 5678
    som_type_copy[np.isin(som_type_copy, s3_array)] = -5
    som_type_copy[np.isin(som_type_copy, s0_array)] = -250
    som_type_copy[som_type_copy == 1234] = 1
    som_type_copy[som_type_copy == 5678] = 2
    som_type_copy[som_type_copy == -5] = 3
    som_type_copy[som_type_copy == -250] = 0

    return som_type_copy


def data_to_log_decile_log_area_aft(peaklet_data, normalization_factor):
    """Converts peaklet data into the current best inputs for the SOM, log10(deciles) + log10(area)
    + AFT Since we are dealing with logs, anything less than 1 will be set to 1."""
    # turn deciles into approriate 'normalized' format
    # (maybe also consider L1 normalization of these inputs)
    decile_data = compute_quantiles(peaklet_data, 10)
    data = peaklet_data.copy()
    decile_data[decile_data < 1] = 1

    decile_log = np.log10(decile_data)
    decile_log_over_max = np.divide(decile_log, normalization_factor[:10])
    # Now lets deal with area
    data["area"] = data["area"] + normalization_factor[11] + 1
    peaklet_log_area = np.log10(data["area"])
    peaklet_aft = (
        np.sum(data["area_per_channel"][:, : straxen.n_top_pmts], axis=1) / peaklet_data["area"]
    )
    peaklet_aft = np.where(peaklet_aft > 0, peaklet_aft, 0)
    peaklet_aft = np.where(peaklet_aft < 1, peaklet_aft, 1)
    deciles_area_aft = np.concatenate(
        (
            decile_log_over_max,
            np.reshape(peaklet_log_area, (len(peaklet_log_area), 1)) / normalization_factor[10],
            np.reshape(peaklet_aft, (len(peaklet_log_area), 1)),
        ),
        axis=1,
    )
    return deciles_area_aft


def compute_quantiles(peaks: np.ndarray, n_samples: int):
    """Compute waveforms and quantiles for a given number of nodes(attributes) :param peaks:

    :param n_samples: number of nodes or attributes
    :return: quantiles

    """

    data = peaks["data"].copy()
    data[data < 0.0] = 0.0
    dt = peaks["dt"]
    q = compute_wf_attributes(data, dt, n_samples)
    return q


@export
@numba.jit(nopython=True, cache=True)
def compute_wf_attributes(data, sample_length, n_samples: int):
    """
    Compute waveform attribures
    Quantiles: represent the amount of time elapsed for
    a given fraction of the total waveform area to be observed in n_samples
    i.e. n_samples = 10, then quantiles are equivalent deciles
    Waveforms: downsampled waveform to n_samples
    :param data: waveform e.g. peaks or peaklets
    :param n_samples: compute quantiles for a given number of samples
    :return: waveforms and quantiles of size n_samples
    """

    assert data.shape[0] == len(sample_length), "ararys must have same size"

    num_samples = data.shape[1]

    quantiles = np.zeros((len(data), n_samples), dtype=np.float64)

    # Cannot compute with with more samples than actual waveform sample
    assert num_samples > n_samples, "cannot compute with more samples than the actual waveform"
    assert num_samples % n_samples == 0, "number of samples must be a multiple of n_samples"

    # Compute quantiles
    inter_points = np.linspace(0.0, 1.0 - (1.0 / n_samples), n_samples)
    cumsum_steps = np.zeros(n_samples + 1, dtype=np.float64)
    frac_of_cumsum = np.zeros(num_samples + 1)
    sample_number_div_dt = np.arange(0, num_samples + 1, 1)
    for i, (samples, dt) in enumerate(zip(data, sample_length)):
        if np.sum(samples) == 0:
            continue
        # reset buffers
        frac_of_cumsum[:] = 0
        cumsum_steps[:] = 0
        frac_of_cumsum[1:] = np.cumsum(samples)
        frac_of_cumsum[1:] = frac_of_cumsum[1:] / frac_of_cumsum[-1]
        cumsum_steps[:-1] = np.interp(inter_points, frac_of_cumsum, sample_number_div_dt * dt)
        cumsum_steps[-1] = sample_number_div_dt[-1] * dt
        quantiles[i] = cumsum_steps[1:] - cumsum_steps[:-1]

    return quantiles



import strax
import straxen
from straxen import pre_apply_function

export, __all__ = strax.exporter()


@export
class EventInfo(strax.MergeOnlyPlugin):
    """Plugin which merges the information of all event data_kinds into a single data_type."""

    depends_on = (
        "event_basics",
        "event_positions",
        "corrected_areas",
        "energy_estimates",
    )
    save_when = strax.SaveWhen.ALWAYS
    provides = "event_info"
    __version__ = "0.0.2"

    event_info_function = straxen.URLConfig(
        default="pre_apply_function",
        infer_type=False,
        help="Function that must be applied to all event_info data. Do not change.",
    )

    def compute(self, **kwargs):
        event_info = super().compute(**kwargs)
        if self.event_info_function != "disabled":
            event_info = pre_apply_function(
                event_info,
                self.run_id,
                self.provides,
                self.event_info_function,
            )
        return event_info



import numpy as np
import strax
import straxen


export, __all__ = strax.exporter()


@export
class PeakBasicsVanilla(strax.Plugin):
    """Compute the basic peak-properties, thereby dropping structured arrays.

    NB: This plugin can therefore be loaded as a pandas DataFrame.

    """

    __version__ = "0.1.7"
    depends_on = "peaks"
    provides = "peak_basics"

    check_peak_sum_area_rtol = straxen.URLConfig(
        default=None,
        track=False,
        infer_type=False,
        help=(
            "Check if the sum area and the sum of area per "
            "channel are the same. If None, don't do the "
            "check. To perform the check, set to the desired "
            " rtol value used e.g. '1e-4' (see np.isclose)."
        ),
    )

    n_top_pmts = straxen.URLConfig(default=straxen.n_top_pmts, type=int, help="Number of top PMTs")

    def infer_dtype(self):
        dtype = strax.time_fields + [
            (("Weighted average center time of the peak [ns]", "center_time"), np.int64),
            (("Peak integral in PE", "area"), np.float32),
            (("Number of hits contributing at least one sample to the peak", "n_hits"), np.int32),
            (("Number of PMTs contributing to the peak", "n_channels"), np.int16),
            (("Number of top PMTs contributing to the peak", "top_n_channels"), np.int16),
            (("PMT number which contributes the most PE", "max_pmt"), np.int16),
            (("Area of signal in the largest-contributing PMT (PE)", "max_pmt_area"), np.float32),
            (("Total number of saturated channels", "n_saturated_channels"), np.int16),
            (("Width (in ns) of the central 50% area of the peak", "range_50p_area"), np.float32),
            (("Width (in ns) of the central 90% area of the peak", "range_90p_area"), np.float32),
            (("Fraction of area seen by the top array", "area_fraction_top"), np.float32),
            (("Length of the peak waveform in samples", "length"), np.int32),
            (("Time resolution of the peak waveform in ns", "dt"), np.int16),
            (("Weighted relative median time of the peak [ns]", "median_time"), np.float32),
            (("Time between 10% and 50% area quantiles [ns]", "rise_time"), np.float32),
            (
                ("Number of PMTs with hits within tight range of mean", "tight_coincidence"),
                np.int16,
            ),
            (("Classification of the peak(let)", "type"), np.int8),
            (("Is merged from peaklets", "merged"), bool),
            (
                ("Largest time difference between apexes of hits inside peak [ns]", "max_diff"),
                np.int32,
            ),
            (
                ("Smallest time difference between apexes of hits inside peak [ns]", "min_diff"),
                np.int32,
            ),
            (
                (
                    "First channel/PMT number inside peak (sorted by apexes of hits)",
                    "first_channel",
                ),
                np.int16,
            ),
            (
                ("Last channel/PMT number inside peak (sorted by apexes of hits)", "last_channel"),
                np.int16,
            ),
        ]
        return dtype

    def compute(self, peaks):
        p = peaks
        r = np.zeros(len(p), self.dtype)
        needed_fields = "time center_time length dt median_time area area_fraction_top type "
        needed_fields += "merged max_diff min_diff first_channel last_channel"
        for q in needed_fields.split():
            r[q] = p[q]
        r["endtime"] = p["time"] + p["dt"] * p["length"]
        r["n_channels"] = (p["area_per_channel"] > 0).sum(axis=1)
        r["top_n_channels"] = (p["area_per_channel"][:, : self.n_top_pmts] > 0).sum(axis=1)
        r["n_hits"] = p["n_hits"]
        r["range_50p_area"] = p["width"][:, 5]
        r["range_90p_area"] = p["width"][:, 9]
        r["max_pmt"] = np.argmax(p["area_per_channel"], axis=1)
        r["max_pmt_area"] = np.max(p["area_per_channel"], axis=1)
        r["tight_coincidence"] = p["tight_coincidence"]
        r["n_saturated_channels"] = p["n_saturated_channels"]
        r["rise_time"] = -p["area_decile_from_midpoint"][:, 1]

        if self.check_peak_sum_area_rtol is not None:
            area_total = p["area_per_channel"].sum(axis=1)
            self.check_area(area_total, p, self.check_peak_sum_area_rtol)
        return r

    @staticmethod
    def check_area(area_per_channel_sum, peaks, rtol) -> None:
        """Check if the area of the sum-wf is the same as the total area (if the area of the peak is
        positively defined).

        :param area_per_channel_sum: the summation of the peaks['area_per_channel'] which will be
            checked against the values of peaks['area'].
        :param peaks: array of peaks.
        :param rtol: relative tolerance for difference between area_per_channel_sum and
            peaks['area']. See np.isclose.
        :raises: ValueError if the peak area and the area-per-channel sum are not sufficiently close

        """
        positive_area = peaks["area"] > 0
        if not np.sum(positive_area):
            return

        is_close = np.isclose(
            area_per_channel_sum[positive_area],
            peaks[positive_area]["area"],
            rtol=rtol,
        )

        if not is_close.all():
            for peak in peaks[positive_area][~is_close]:
                print("bad area")
                strax.print_record(peak)

            p_i = np.where(~is_close)[0][0]
            peak = peaks[positive_area][p_i]
            area_fraction_off = 1 - area_per_channel_sum[positive_area][p_i] / peak["area"]
            message = (
                "Area not calculated correctly, it's "
                f"{100 * area_fraction_off} % off, time: {peak['time']}"
            )
            raise ValueError(message)



import strax
from ._peak_positions_base import PeakPositionsBase

export, __all__ = strax.exporter()


@export
class PeakPositionsMLP(PeakPositionsBase):

    __version__ = "0.0.0"
    child_plugin = True
    algorithm = "mlp"
    depends_on = (
        "peaklet_positions_mlp",
        "enhanced_peaklet_classification",
        "merged_s2s",
        "merged_s2_positions_mlp",
    )
    provides = "peak_positions_mlp"



import numpy as np
import straxen
import strax
import numba

export, __all__ = strax.exporter()
__all__.extend(["DTYPE_VETO_INTERVALS"])


DTYPE_VETO_INTERVALS = np.int64


@export
class VetoProximity(strax.ExhaustPlugin):
    """Find the closest next/previous veto start w.r.t.

    the event time or when a busy happens during an event.

    """

    __version__ = "2.2.0"
    # Strictly speaking, we could depend on 'events', but then you couldn't
    # change the event_window_fields to e.g. s1_time and s2_endtime.
    depends_on = ("event_basics", "veto_intervals")
    provides = "veto_proximity"
    data_kind = "events"

    event_window_fields = straxen.URLConfig(
        default=("time", "endtime"),
        help=(
            "Fields to determine where to look for overlaps for using "
            "this plugin in the events. The default uses start and endtime "
            "of an event, but this can also be the S1 or S2 start/endtime"
        ),
    )

    veto_names = ["busy", "busy_he", "hev", "straxen_deadtime"]

    def infer_dtype(self):
        dtype = []
        dtype += strax.time_fields
        start_field, stop_field = self.event_window_fields
        for name in self.veto_names:
            dtype += [
                (
                    (
                        f'Duration of event overlapping with "{name}"-veto [ns]',
                        f"veto_{name}_overlap",
                    ),
                    DTYPE_VETO_INTERVALS,
                ),
                (
                    (
                        (
                            f'Time (absolute value) to previous "{name}"-veto '
                            f'from "{start_field}" of event [ns]'
                        ),
                        f"time_to_previous_{name}",
                    ),
                    DTYPE_VETO_INTERVALS,
                ),
                (
                    (
                        (
                            f'Time (absolute value) to next "{name}"-veto '
                            f'from "{stop_field}" of event [ns]'
                        ),
                        f"time_to_next_{name}",
                    ),
                    DTYPE_VETO_INTERVALS,
                ),
            ]

        return dtype

    def set_result_for_veto(
        self,
        result_buffer: np.ndarray,
        event_window: np.ndarray,
        veto_intervals: np.ndarray,
        veto_name: str,
    ) -> None:
        """Fill the result buffer inplace. Goal is to find vetos with <veto_name> that are either
        during, before or after the current event_window.

        :param result_buffer: The buffer to fill inplace
        :param event_window: start/stop boundaries of the event to consider. Should be an array with
            ['time'] and ['endtime'] which can be based on event start/end times or S1/S2 times
        :param veto_intervals: veto intervals datatype
        :param veto_name: The name of the veto to fill the result buffer for
        :return: Nothing, results are filled in place

        """
        # Set defaults to be some very long time
        result_buffer[f"time_to_previous_{veto_name}"] = np.iinfo(DTYPE_VETO_INTERVALS).max
        result_buffer[f"time_to_next_{veto_name}"] = np.iinfo(DTYPE_VETO_INTERVALS).max

        selected_intervals = veto_intervals[veto_intervals["veto_type"] == f"{veto_name}_veto"]
        if not len(selected_intervals):
            return

        vetos_during_event = strax.touching_windows(selected_intervals, event_window)

        # Figure out the vetos *during* an event
        res = self.get_overlapping_window_time(
            vetos_during_event, selected_intervals, event_window, result_buffer
        )
        result_buffer[f"veto_{veto_name}_overlap"] = res

        # Find the next and previous veto's
        times_to_prev, times_to_next = strax.abs_time_to_prev_next_interval(
            event_window, selected_intervals
        )
        mask_prev = times_to_prev > 0
        result_buffer[f"time_to_previous_{veto_name}"][mask_prev] = times_to_prev[mask_prev]

        max_next = times_to_next > 0
        result_buffer[f"time_to_next_{veto_name}"][max_next] = times_to_next[max_next]

    @staticmethod
    @numba.njit
    def get_overlapping_window_time(
        vetos_during_event, selected_intervals, event_window, result_buffer
    ):
        """Computes total time each event overlaps with the corresponding veto."""
        res = np.zeros(len(vetos_during_event), DTYPE_VETO_INTERVALS)

        for event_i, veto_window in enumerate(vetos_during_event):
            if veto_window[1] - veto_window[0]:
                vetos_in_window = selected_intervals[veto_window[0] : veto_window[1]].copy()
                starts = np.clip(
                    vetos_in_window["time"],
                    event_window[event_i]["time"],
                    event_window[event_i]["endtime"],
                )
                stops = np.clip(
                    vetos_in_window["endtime"],
                    event_window[event_i]["time"],
                    event_window[event_i]["endtime"],
                )

                res[event_i] = np.sum(stops - starts)
        return res

    def compute(self, events, veto_intervals):
        result = np.zeros(len(events), self.dtype)
        result["time"] = events["time"]
        result["endtime"] = events["endtime"]

        # Get containers for touching windows based on self.event_window_fields
        event_window = np.zeros(len(events), dtype=strax.time_fields)
        event_window["time"] = events[self.event_window_fields[0]]
        event_window["endtime"] = events[self.event_window_fields[1]]

        for veto_name in self.veto_names:
            self.set_result_for_veto(result, event_window, veto_intervals, veto_name)
        return result



from .regex_dispatcher import RegexDispatcher

import straxen
import re
import warnings

check_urls = RegexDispatcher("check_urls")


class URLWarning(UserWarning):
    pass


@check_urls.register(r"(.*)(.*cmt.*)")
def not_cmt_check(url: str):
    raise NotImplementedError("Error, the cmt protocol is removed. Please use xedocs instead.")


@check_urls.register(r"(.*)(.*xedocs.*)")
def confirm_xedocs_schema_exists_check(url: str):
    import xedocs

    # Selects the schema name in the URL and confirms it exists in the xedocs db
    if not re.findall(r"(?<=\://)[^\://]*(?=\?)", url)[0] in xedocs.list_schemas():
        warnings.warn(
            f"You tried to use a schema that is not part of the xedocs database. url: {url}",
            URLWarning,
        )


# @check_urls.register(r".*(list-to-array|'objects-to-dict').*")
@check_urls.register(r"(.*)(.*xedocs.*)")
def data_as_list_check(url: str):
    if ("list-to-array" in url) or ("objects-to-dict" in url):
        if not straxen.URLConfig.kwarg_from_url(url, "as_list"):
            warnings.warn(
                "When using the list-to-array or objects-to-dict protocol, "
                f"you must include an as_list=True in the URL arguments. url: {url}",
                URLWarning,
            )
        if not ("sort" in url):
            warnings.warn(
                "When using the list-to-array or objects-to-dict protocol, "
                f"you must include a sort argument in the URL. url: {url}",
                URLWarning,
            )


@check_urls.register(r"(.*)(.*xedocs.*)")
def are_resources_needed_check(url: str):
    parent_class = get_parent_class(url)

    if ("BaseMap" in parent_class) or ("BaseResourceReference" in parent_class):
        if not ("resource://" in url) and straxen.URLConfig.kwarg_from_url(url, "attr") != "map":
            warnings.warn(
                "A URL which requests the resource:// was given. However resource:// "
                f"was not found within the URL, not data will be loaded. url: {url}",
                URLWarning,
            )


@check_urls.register(r"(.*)(.*xedocs.*)")
def itp_map_check(url: str):
    parent_class = get_parent_class(url)

    if "BaseMap" in parent_class:
        if (
            not ("itp_map://" in url)
            and straxen.URLConfig.kwarg_from_url(url, "attr") != "map"
            and not ("tf://" in url)
        ):
            warnings.warn(
                "Warning, you are requesting a map file with this URL. However, the protocol "
                "itp_map:// was not requested as part of the URL. "
                f"The itp_map:// protocol is requiered for map corrections. url: {url}",
                URLWarning,
            )


@check_urls.register(r".*posrec_models.*")
def posrec_models_check(url: str):
    if not ("tf://" in url) and not ("jax://" in url):
        warnings.warn(
            "Warning, you are requesting a position reconstruction model with this URL. However, "
            "the protocol tf:// and jax:// were not requested as part of the URL. "
            "The tf:// or jax://protocol is requiered for position reconstruction corrections. "
            f"url: {url}",
            URLWarning,
        )


@check_urls.register(r"(.*)(.*xedocs.*)")
def url_attr_check(url: str):
    if not ("attr" in url):
        warnings.warn(
            "A URL without an 'attr' argument was given, as a result, instead of a value, "
            "list of values or other files, you will get a document, "
            f"which cannot be used to process data. url: {url}",
            URLWarning,
        )


@check_urls.register(r"(.*)(.*xedocs.*)")
def url_version_check(url: str):
    if not ("version" in url):
        warnings.warn(
            "A URL without a 'version' argument was given, as a result, to use a url protocol "
            f"to get a correction a version of said correcection is requiered. url: {url}",
            URLWarning,
        )


@check_urls.register(r".*objects-to-dict.*")
def dict_attributes(url: str):
    if not ("key_attr" in url):
        warnings.warn(
            "Warning, you are requesting a correction in the form of a dictionary with "
            "this URL. However, you did not choose keys for the dictionary with [key_attr]. "
            f"Please insert [key_attr=] in your url for the keys. url: {url}",
            URLWarning,
        )
    if not ("value_attr" in url):
        warnings.warn(
            "Warning, you are requesting a correction in the form of a dictionary with "
            "this URL. However, you did not include [value_attr] in your url which is needed "
            "for dictionry outputs. Please insert [key_attr=] "
            f"in your url for the keys. url: {url}",
            URLWarning,
        )


def get_parent_class(url: str):
    """Finds a xedocs schema and returns the class used to make the schema."""
    import xedocs

    schema = re.findall(r"(?<=\://)[^\://]*(?=\?)", url)[0]
    xedocs_class = xedocs.find_schema(schema).__mro__
    parent_class = [parent.__name__ for parent in xedocs_class]

    return parent_class



# mypy: disable-error-code="no-redef"
__version__ = "3.2.2"

from utilix import uconfig
from .common import *

from .itp_map import *
from .matplotlib_utils import *
from .mini_analysis import *
from .misc import *

from .scada import *
from .bokeh_utils import *
from .config.url_config import *

from . import plugins
from .plugins import *

from . import storage
from .storage import *

from . import analyses

from . import config

from . import units

# Do not make all contexts directly available under straxen.
# Otherwise, we have straxen.demo() etc.
from . import contexts

from . import test_utils
from .test_utils import *

from . import docs_utils
from .docs_utils import *

from . import daq_core

try:
    from . import holoviews_utils
    from .holoviews_utils import *
except ModuleNotFoundError:
    pass

from .entry_points import load_entry_points

load_entry_points()
del load_entry_points



import param
import panel as pn
import holoviews as hv
import numpy as np
from immutabledict import immutabledict


class EventStats(param.Parameterized):
    """Class which displays all events in a user defined data space. Marks position of single
    selected event depending on the passed index or index widget.

    :param events: Events for which histogram should be plotted.
    :param default_config: Dictionary containing the default settings
        for the plot.
    :param exclude_fields: A list of fields to be excluded from the
        x/y-selection widget. Can be uselful as event_info has a lot of
        fields. Can be either a list/tuple or a function. If passed as
        function must take as an argument a list with field names and
        return a list with field names.
    :param _index_slider: Parameter to pass an external integer slider
        to class. E.g. required in event display.

    Warning:
        In case the initally selected value is outside of the specified
        plotting range the plot will break.

    """

    _x_edges = None
    _y_edges = None
    _entries = None
    _x_field_name = None
    _y_field_name = None
    _x_bins = None
    _y_bins = None
    _x_log = False
    _y_log = False
    _z_log = False

    # Static widgets used for setting plot layout.
    x_field = "cs1"
    x_bins = param.Tuple(default=(200, 0, 50), label="x binning (#bins, min max)")
    x_log = param.Boolean(default=False, label="Logarithmic x binning (np.logspace)")
    y_field = "cs2"
    y_bins = param.Tuple(default=(200, 1000, 2500), label="y binning (#bins, min max)")
    y_log = param.Boolean(default=False, label="Logarithmic y binning (np.logspace)")
    clim = param.Tuple(default=(1, None), label="Colorbarlimit (min, max)")
    z_log = param.Boolean(default=True, label="Logarithmic colorbar")
    VALID_CONFIGS = ["x_field", "x_bins", "x_log", "y_field", "y_bins", "y_log", "clim", "z_log"]

    def __init__(
        self,
        events,
        default_config=immutabledict(),
        exclude_fields=None,
        _index_slider=None,
        **params,
    ):
        for key, value in default_config.items():
            if not hasattr(self, key):
                raise ValueError(
                    f'"{key}" is not a valid config. '
                    f"Only the following attributes are recognized: {self.VALID_CONFIGS}."
                )
            setattr(self, key, value)

        self.exclude_fields = exclude_fields
        self.events = events
        self._create_widgets(
            self.x_field,
            self.y_field,
            _index_slider,
        )
        super().__init__(**params)

    def interactive_event_stats(self):
        """An interactive plot which displays a histogram of the supplied data.

        The user can manually change the axis and binning via the provided widgets. This plot is
        meant to be used as a stand alone outside of the event display.

        """
        stream_hist, stream_marker = self._create_streams()
        hist = hv.DynamicMap(self._plot_hist_data, streams=stream_hist)
        selected_points = hv.DynamicMap(self._plot_selected_event, streams=stream_marker)
        return pn.Column(
            (hist * selected_points).opts(width=400),
            self._make_widget_panel(include_index_slider=True),
        )

    def _create_streams(self):
        """Creates stream objects for the histogram and marker plot."""
        stream_hist = dict(
            x_field=self.x_field_widget,
            y_field=self.y_field_widget,
            x_bins=self.param.x_bins,
            y_bins=self.param.y_bins,
            clim=self.param.clim,
            x_log=self.param.x_log,
            y_log=self.param.y_log,
            z_log=self.param.z_log,
        )

        stream_marker = dict(
            x_field=self.x_field_widget, y_field=self.y_field_widget, index=self.index_slider
        )

        return stream_hist, stream_marker

    def _create_widgets(
        self,
        x_field="cs1",
        y_field="cs2",
        _index_slider=None,
    ):
        """Creates widgets for the x and y field as well as an slider for the event index if not
        supplied externally."""
        fields = list(self.events.dtype.names)
        if self.exclude_fields:
            if hasattr(self.exclude_fields, "__call__"):
                fields = self.exclude_fields(fields)
            elif isinstance(self.exclude_fields, (list, tuple)):
                fields = self._exclude_fields(fields, self.exclude_fields)
            else:
                raise ValueError('"exclude_fields" must be either a list or a function!')

        fields.sort()
        self.x_field_widget = pn.widgets.Select(value=x_field, options=fields, name="X-field")

        self.y_field_widget = pn.widgets.Select(value=y_field, options=fields, name="Y-field")

        if not _index_slider:
            self.index_slider = pn.widgets.IntSlider(
                start=0, end=len(self.events), name="Event Index:"
            )
        else:
            self.index_slider = _index_slider

    @staticmethod
    def _exclude_fields(fields, excluded_fields):
        """Function which drops certain fields from."""
        dtype = []
        for f in fields:
            if f not in excluded_fields:
                dtype.append(f)
        return dtype

    def _make_widget_panel(self, include_index_slider=False):
        """Put widgets used in the selection plot into a single panel which can be displayed.

        Adds index slider for stand alone plot.

        """
        widget_panel = pn.Row(
            pn.Column(
                self.x_field_widget,
                self.param.x_bins,
                self.param.x_log,
            ),
            pn.Column(
                self.y_field_widget,
                self.param.y_bins,
                self.param.y_log,
            ),
            pn.Column(
                self.param.clim,
                self.param.z_log,
            ),
            scroll=True,
        )

        if include_index_slider:
            widget_panel = pn.Column(self.index_slider, widget_panel)

        return widget_panel

    def _hist_data(self, x_field, y_field, xbins, ybins, x_log, y_log):
        """Function which checks if underlying data has to be rebinned."""
        _hist_has_not_changed = True
        _hist_has_not_changed &= x_field == self._x_field_name
        _hist_has_not_changed &= y_field == self._y_field_name
        _hist_has_not_changed &= xbins == self._x_bins
        _hist_has_not_changed &= x_log == self._x_log
        _hist_has_not_changed &= ybins == self._y_bins
        _hist_has_not_changed &= y_log == self._y_log

        if not _hist_has_not_changed:
            x_binning = self._get_binning(*xbins, x_log)
            y_binning = self._get_binning(*ybins, y_log)

            entries, x_edges, y_edges = np.histogram2d(
                self.events[x_field],
                self.events[y_field],
                bins=(
                    x_binning,
                    y_binning,
                ),
            )

            self._x_edges = x_edges
            self._y_edges = y_edges
            self._entries = entries.T

            self._x_field_name = x_field
            self._y_field_name = y_field
            self._x_bins = xbins
            self._y_bins = ybins
            self._x_log = x_log
            self._y_log = y_log

    @staticmethod
    def _get_binning(n_bins, bin_min, bin_max, log):
        """Function which will return an array of bin edges."""
        if log:
            bin_min = max(10**-3, bin_min)
            bins = np.logspace(np.log10(bin_min), np.log10(bin_max), n_bins)
        else:
            bins = np.linspace(bin_min, bin_max, n_bins)
        return bins

    def _plot_hist_data(
        self,
        x_field: str,
        y_field: str,
        x_bins: tuple,
        y_bins: tuple,
        clim: tuple = (1, None),
        x_log: bool = False,
        y_log: bool = False,
        z_log: bool = True,
    ):
        """Creates histogram based on user specified settings."""
        self._hist_data(
            x_field,
            y_field,
            x_bins,
            y_bins,
            x_log,
            y_log,
        )
        hist = hv.QuadMesh(
            (self._x_edges, self._y_edges, self._entries),
            [x_field, y_field],
        )
        hist = hist.apply.opts(
            logz=z_log,
            logx=x_log,
            logy=y_log,
            clim=clim,
            cmap="viridis",
            colorbar=True,
            framewise=True,
            clipping_colors={"min": "white"},
        )
        return hist

    def _plot_selected_event(self, x_field: str, y_field: str, index: int = 0):
        """Function which highlights event with selected index in plot."""
        point_x = self.events[x_field][index]
        point_y = self.events[y_field][index]
        _is_nan = np.isnan(point_x) or np.isnan(point_y)
        if not _is_nan:
            selected_points = hv.Points(
                (point_x, point_y), [self.x_field_widget.value, self.y_field_widget.value]
            ).opts(color="red", size=7, alpha=1)

            hline = hv.HLine(point_y).opts(color="red", alpha=0.4)
            vline = hv.VLine(point_x).opts(color="red", alpha=0.4)
        else:
            # In case one of the parameters is not defined
            # plot invisilbe point
            selected_points = hv.Points(
                (0, 0), [self.x_field_widget.value, self.y_field_widget.value]
            ).opts(color="k", alpha=0, size=7)

            hline = hv.HLine(0).opts(color="k", alpha=0)
            vline = hv.VLine(0).opts(color="k", alpha=0)

        return selected_points * hline * vline



import os
import warnings
from typing import Dict, Any, List, Optional
from immutabledict import immutabledict
import socket

import strax
import straxen

from straxen import HAVE_ADMIX

common_opts: Dict[str, Any] = dict(
    register_all=[straxen.plugins],
    register=[
        straxen.PulseProcessing,
        straxen.Peaklets,
        straxen.PeakletClassificationSOM,
        straxen.MergedS2s,
        straxen.PeaksSOM,
        straxen.PeakBasicsSOM,
        straxen.PeakProximity,
        straxen.Events,
        straxen.EventBasicsSOM,
        straxen.EventPositions,
        straxen.CorrectedAreas,
        straxen.EnergyEstimates,
        straxen.EventInfoDouble,
        straxen.DistinctChannels,
        straxen.PeakPositionsMLP,
        straxen.PeakPositionsCNF,
    ],
    check_available=("peak_basics", "event_basics"),
    store_run_fields=("name", "number", "start", "end", "livetime", "mode", "source"),
    use_per_run_defaults=False,
)


common_config = dict(
    n_tpc_pmts=straxen.n_tpc_pmts,
    n_top_pmts=straxen.n_top_pmts,
    gain_model=(
        "list-to-array://xedocs://pmt_area_to_pes"
        "?as_list=True&sort=pmt&detector=tpc"
        "&run_id=plugin.run_id&version=ONLINE&attr=value"
    ),
    gain_model_nv=(
        "list-to-array://xedocs://pmt_area_to_pes"
        "?as_list=True&sort=pmt&detector=neutron_veto"
        "&run_id=plugin.run_id&version=ONLINE&attr=value"
    ),
    gain_model_mv=(
        "list-to-array://xedocs://pmt_area_to_pes"
        "?as_list=True&sort=pmt&detector=muon_veto"
        "&run_id=plugin.run_id&version=ONLINE&attr=value"
    ),
    channel_map=immutabledict(
        # (Minimum channel, maximum channel)
        # Channels must be listed in a ascending order!
        tpc=(0, 493),
        he=(500, 752),  # high energy
        aqmon=(790, 807),
        aqmon_nv=(808, 815),  # nveto acquisition monitor
        tpc_blank=(999, 999),
        mv=(1000, 1083),
        aux_mv=(1084, 1087),  # Aux mv channel 2 empty 1 pulser and 1 GPS
        mv_blank=(1999, 1999),
        nveto=(2000, 2119),
        nveto_blank=(2999, 2999),
    ),
)


def find_rucio_local_path(include_rucio_local, _rucio_local_path):
    """Check the hostname to determine which rucio local path to use. Note that access to
    /dali/lgrandi/rucio/ is possible only if you are on dali compute node or login node.

    :param include_rucio_local: add the rucio local storage frontend. This is only needed if one
        wants to do a fuzzy search in the data the runs database is out of sync with rucio
    :param _rucio_local_path: str, path of local RSE of rucio. Only use for testing!

    """
    hostname = socket.gethostname()
    # if you are on dali compute node, do nothing
    if ("dali" in hostname) and ("login" not in hostname):
        _include_rucio_local = include_rucio_local
        __rucio_local_path = _rucio_local_path
    # Assumed the only other option is 'midway' or login nodes,
    # where we have full access to dali and project space.
    # This doesn't make sense outside XENON but we don't care.
    else:
        _include_rucio_local = True
        __rucio_local_path = "/project/lgrandi/rucio/"
        print(
            "You specified _auto_append_rucio_local=True and you are not on dali compute nodes, "
            f"so we will add the following rucio local path: {__rucio_local_path}"
        )

    return _include_rucio_local, __rucio_local_path


def xenonnt(
    output_folder: str = "./strax_data",
    we_are_the_daq: bool = False,
    minimum_run_number: int = 7157,
    maximum_run_number: Optional[int] = None,
    # Frontends
    include_rucio_remote: bool = False,
    include_online_monitor: bool = False,
    include_rucio_local: bool = False,
    # Frontend options
    download_heavy: bool = False,
    remove_heavy: bool = False,
    _auto_append_rucio_local: bool = True,
    _rucio_path: str = "/dali/lgrandi/rucio/",
    _rucio_local_path: Optional[str] = None,
    _raw_paths: List[str] = ["/dali/lgrandi/xenonnt/raw"],
    _processed_paths: List[str] = [
        "/project/lgrandi/xenonnt/processed",
        "/project/lgrandi/xenonnt/processed_sr2_offline_round_1",
        "/project/lgrandi/xenonnt/processed_sr2_offline_round_2",
        "/project/lgrandi/xenonnt/processed_sr2_offline_round_3",
        "/project2/lgrandi/xenonnt/processed",
        "/dali/lgrandi/xenonnt/processed",
        "/dali/lgrandi/xenonnt/processed_sr2_offline_round_1",
        "/dali/lgrandi/xenonnt/processed_sr2_offline_round_2",
        "/dali/lgrandi/xenonnt/processed_sr2_offline_round_3",
    ],
    # Testing options
    _database_init: bool = True,
    _forbid_creation_of: Optional[dict] = None,
    **kwargs,
):
    """XENONnT online processing and analysis.

    :param output_folder: str, Path of the strax.DataDirectory where new data can be stored
    :param we_are_the_daq: bool, if we have admin access to upload data
    :param minimum_run_number: int, lowest number to consider
    :param maximum_run_number: Highest number to consider. When None (the default) consider all runs
        that are higher than the minimum_run_number.
    :param include_rucio_remote: add the rucio remote frontend to the context
    :param include_online_monitor: add the online monitor storage frontend
    :param include_rucio_local: add the rucio local storage frontend. This is only needed if one
        wants to do a fuzzy search in the data the runs database is out of sync with rucio
    :param download_heavy: bool, whether or not to allow downloads of heavy data (raw_records*, less
        the aqmon)
    :param remove_heavy: bool, whether or not to remove the heavy data after reading
    :param _auto_append_rucio_local: bool, whether or not to automatically append the rucio local
        path
    :param _rucio_path: str, path of rucio
    :param _rucio_local_path: str, path of local RSE of rucio. Only use for testing!
    :param _raw_paths: list[str], common path of the raw-data
    :param _processed_paths: list[str]. common paths of output data
    :param _database_init: bool, start the database (for testing)
    :param _forbid_creation_of: str/tuple, of datatypes to prevent form being written (raw_records*
        is always forbidden).
    :param kwargs: dict, context options
    :return: strax.Context

    """
    if "xedocs_version" in kwargs:
        warnings.warn(
            "Please use xenonnt_* instead of xenonnt if you want to specify xedocs_version"
        )

    context_options = {**straxen.contexts.common_opts, **kwargs}

    st = strax.Context(config=straxen.contexts.common_config, **context_options)
    st.register(
        [
            straxen.DAQReader,
            straxen.LEDCalibration,
            straxen.LEDAfterpulseProcessing,
            straxen.nVeto_reflectivity,
        ]
    )

    if _auto_append_rucio_local:
        include_rucio_local, _rucio_local_path = find_rucio_local_path(
            include_rucio_local, _rucio_local_path
        )

    st.storage = (
        [
            straxen.RunDB(
                readonly=not we_are_the_daq,
                minimum_run_number=minimum_run_number,
                maximum_run_number=maximum_run_number,
                runid_field="number",
                new_data_path=output_folder,
                rucio_path=_rucio_path,
            )
        ]
        if _database_init
        else []
    )
    if not we_are_the_daq:
        for _raw_path in _raw_paths:
            st.storage += [
                strax.DataDirectory(_raw_path, readonly=True, take_only=straxen.DAQReader.provides)
            ]
        for _processed_path in _processed_paths:
            st.storage += [strax.DataDirectory(_processed_path, readonly=True)]

        if output_folder:
            st.storage += [
                strax.DataDirectory(
                    output_folder,
                    provide_run_metadata=True,
                )
            ]
        st.context_config["forbid_creation_of"] = straxen.daqreader.DAQReader.provides
        if _forbid_creation_of is not None:
            st.context_config["forbid_creation_of"] += strax.to_str_tuple(_forbid_creation_of)

    # Add the rucio frontend if we are able to
    if include_rucio_remote and HAVE_ADMIX:
        rucio_frontend = straxen.RucioRemoteFrontend(
            staging_dir=os.path.join(output_folder, "rucio"),
            download_heavy=download_heavy,
            remove_heavy=remove_heavy,
        )
        st.storage += [rucio_frontend]

    if include_rucio_local:
        rucio_local_frontend = straxen.RucioLocalFrontend(path=_rucio_local_path)
        st.storage += [rucio_local_frontend]

    # Only the online monitor backend for the DAQ
    if _database_init and (include_online_monitor or we_are_the_daq):
        st.storage += [
            straxen.OnlineMonitor(
                readonly=not we_are_the_daq,
                take_only=(
                    "veto_intervals",
                    "online_peak_monitor",
                    "event_basics",
                    "online_monitor_nv",
                    "online_monitor_mv",
                    "individual_peak_monitor",
                ),
            )
        ]

    # Remap the data if it is before channel swap (because of wrongly cabled
    # signal cable connectors) These are runs older than run 8797. Runs
    # newer than 8796 are not affected. See:
    # https://github.com/XENONnT/straxen/pull/166 and
    # https://xe1t-wiki.lngs.infn.it/doku.php?id=xenon:xenonnt:dsg:daq:sector_swap
    st.set_context_config(
        {
            "apply_data_function": (
                straxen.remap_old,
                straxen.check_loading_allowed,
            )
        }
    )

    return st


if "xedocs_version" not in strax.Context.takes_config:
    strax.Context = strax.takes_config(
        strax.Option(
            name="xedocs_version",
            default=None,
            type=str,
            help="The version of the xedocs database to use",
        ),
    )(strax.Context)


@strax.Context.add_method
def apply_xedocs_configs(context: strax.Context, db="straxen_db", **kwargs) -> None:
    import xedocs

    if isinstance(db, str):
        func = getattr(xedocs.databases, db)
        db_kwargs = straxen.filter_kwargs(func, kwargs)
        db = func(**db_kwargs)

    filter_kwargs = {k: v for k, v in kwargs.items() if k in db.context_configs.schema.__fields__}

    docs = db.context_configs.find_docs(**filter_kwargs)

    global_config = {doc.config_name: doc.value for doc in docs}

    if len(global_config):
        context.set_config(global_config)
        context.set_context_config({"xedocs_version": filter_kwargs["version"]})
    else:
        raise KeyError(
            f"Could not find any context configs matching {filter_kwargs} in the xedocs database"
        )


def xenonnt_online(xedocs_version="global_ONLINE", _from_cutax=False, **kwargs):
    """XENONnT context."""
    if not _from_cutax and xedocs_version != "global_ONLINE":
        warnings.warn("Don't load a context directly from straxen, use cutax instead!")

    st = straxen.contexts.xenonnt(**kwargs)
    st.apply_xedocs_configs(version=xedocs_version, **kwargs)

    return st


def xenonnt_led(**kwargs):
    st = xenonnt_online(**kwargs)
    st.set_context_config(
        {
            "check_available": ("raw_records", "led_calibration"),
            "free_options": list(common_config.keys()),
        }
    )
    # Return a new context with only raw_records and led_calibration registered
    st = st.new_context(replace=True, config=st.config, storage=st.storage, **st.context_config)
    st.register(
        [
            straxen.DAQReader,
            straxen.LEDCalibration,
            straxen.nVETORecorder,
            straxen.nVETOPulseProcessing,
            straxen.nVETOHitlets,
            straxen.nVetoExtTimings,
        ]
    )
    st.set_config({"coincidence_level_recorder_nv": 1})
    return st
